@@ -228,6 +228,55 @@ where
             compiler,
         })
     }
+
+    fn extract_rocm_arg(args: &ParsedArguments, flag: &str) -> Option<PathBuf> {
+        args.common_args.iter().find_map(|arg| match arg.to_str() {
+            Some(sarg) if sarg.starts_with(flag) => {
+                Some(PathBuf::from(sarg[arg.len()..].to_string()))
+            }
+            _ => None,
+        })
+    }
+
+    fn extract_rocm_env(env_vars: &[(OsString, OsString)], name: &str) -> Option<PathBuf> {
+        env_vars.iter().find_map(|(k, v)| match v.to_str() {
+            Some(path) if k == name => Some(PathBuf::from(path.to_string())),
+            _ => None,
+        })
+    }
+
+    // See https://clang.llvm.org/docs/HIPSupport.html for details regarding the
+    // order in which the environment variables and command-line arguments control the
+    // directory to search for bitcode libraries.
+    fn search_hip_device_libs(
+        args: &ParsedArguments,
+        env_vars: &[(OsString, OsString)],
+    ) -> Vec<PathBuf> {
+        let rocm_path_arg: Option<PathBuf> = Self::extract_rocm_arg(args, "--rocm-path=");
+        let hip_device_lib_path_arg: Option<PathBuf> =
+            Self::extract_rocm_arg(args, "--hip-device-lib-path=");
+        let rocm_path_env: Option<PathBuf> = Self::extract_rocm_env(env_vars, "ROCM_PATH");
+        let hip_device_lib_path_env: Option<PathBuf> =
+            Self::extract_rocm_env(env_vars, "HIP_DEVICE_LIB_PATH");
+
+        let hip_device_lib_path: PathBuf = hip_device_lib_path_arg
+            .or(hip_device_lib_path_env)
+            .or(rocm_path_arg.map(|path| path.join("amdgcn").join("bitcode")))
+            .or(rocm_path_env.map(|path| path.join("amdgcn").join("bitcode")))
+            // This is the default location in official AMD packages and containers.
+            .unwrap_or(PathBuf::from("/opt/rocm/amdgcn/bitcode"));
+
+        hip_device_lib_path
+            .read_dir()
+            .ok()
+            .map(|f| {
+                f.flatten()
+                    .filter(|f| f.path().extension().map_or(false, |ext| ext == "bc"))
+                    .map(|f| f.path())
+                    .collect()
+            })
+            .unwrap_or_default()
+    }
 }
 
 impl<T: CommandCreatorSync, I: CCompilerImpl> Compiler<T> for CCompiler<I> {
@@ -249,11 +298,29 @@ impl<T: CommandCreatorSync, I: CCompilerImpl> Compiler<T> for CCompiler<I> {
     ) -> CompilerArguments<Box<dyn CompilerHasher<T> + 'static>> {
         match self.compiler.parse_arguments(arguments, cwd) {
             CompilerArguments::Ok(mut args) => {
+                // Handle SCCACHE_EXTRAFILES
                 for (k, v) in env_vars.iter() {
                     if k.as_os_str() == OsStr::new("SCCACHE_EXTRAFILES") {
                         args.extra_hash_files.extend(std::env::split_paths(&v))
                     }
                 }
+
+                // Handle cache invalidation for the ROCm device bitcode libraries. Every HIP
+                // object links in some LLVM bitcode libraries (.bc files), so in some sense
+                // every HIP object compilation has an direct dependency on those bitcode
+                // libraries.
+                //
+                // The bitcode libraries are unlikely to change **except** when a ROCm version
+                // changes, so for correctness we should take these bitcode libraries into
+                // account by adding them to `extra_hash_files`.
+                //
+                // In reality, not every available bitcode library is needed, but that is
+                // too much to handle on our side so we just hash every bitcode library we find.
+                if args.language == Language::Hip {
+                    args.extra_hash_files
+                        .extend(Self::search_hip_device_libs(&args, env_vars))
+                }
+
                 CompilerArguments::Ok(Box::new(CCompilerHasher {
                     parsed_args: args,
                     executable: self.executable.clone(),
@@ -315,24 +382,27 @@ where
         // Try to look for a cached preprocessing step for this compilation
         // request.
         let preprocessor_cache_mode_config = storage.preprocessor_cache_mode_config();
-        let mut preprocessor_key = if preprocessor_cache_mode_config.use_preprocessor_cache_mode {
-            preprocessor_cache_entry_hash_key(
-                &executable_digest,
-                parsed_args.language,
-                &preprocessor_and_arch_args,
-                &extra_hashes,
-                &env_vars,
-                &absolute_input_path,
-                compiler.plusplus(),
-                preprocessor_cache_mode_config,
-            )?
-        } else {
-            None
-        };
+        // Disable preprocessor cache when doing distributed compilation
+        let mut preprocessor_key =
+            if !may_dist && preprocessor_cache_mode_config.use_preprocessor_cache_mode {
+                preprocessor_cache_entry_hash_key(
+                    &executable_digest,
+                    parsed_args.language,
+                    &preprocessor_and_arch_args,
+                    &extra_hashes,
+                    &env_vars,
+                    &absolute_input_path,
+                    compiler.plusplus(),
+                    preprocessor_cache_mode_config,
+                )?
+            } else {
+                None
+            };
         if let Some(preprocessor_key) = &preprocessor_key {
             if cache_control == CacheControl::Default {
-                if let Some(mut seekable) =
-                    storage.get_preprocessor_cache_entry(preprocessor_key)?
+                if let Some(mut seekable) = storage
+                    .get_preprocessor_cache_entry(preprocessor_key)
+                    .await?
                 {
                     let mut buf = vec![];
                     seekable.read_to_end(&mut buf)?;
@@ -349,10 +419,13 @@ where
                             "Preprocessor cache updated because of time macros: {preprocessor_key}"
                         );
 
-                        if let Err(e) = storage.put_preprocessor_cache_entry(
-                            preprocessor_key,
-                            preprocessor_cache_entry,
-                        ) {
+                        if let Err(e) = storage
+                            .put_preprocessor_cache_entry(
+                                preprocessor_key,
+                                preprocessor_cache_entry,
+                            )
+                            .await
+                        {
                             debug!("Failed to update preprocessor cache: {}", e);
                             update_failed = true;
                         }
@@ -416,16 +489,14 @@ where
             debug!("removing files {:?}", &outputs);
 
             let v: std::result::Result<(), std::io::Error> =
-                outputs.values().fold(Ok(()), |r, output| {
-                    r.and_then(|_| {
-                        let mut path = args_cwd.clone();
-                        path.push(&output.path);
-                        match fs::metadata(&path) {
-                            // File exists, remove it.
-                            Ok(_) => fs::remove_file(&path),
-                            _ => Ok(()),
-                        }
-                    })
+                outputs.values().try_for_each(|output| {
+                    let mut path = args_cwd.clone();
+                    path.push(&output.path);
+                    match fs::metadata(&path) {
+                        // File exists, remove it.
+                        Ok(_) => fs::remove_file(&path),
+                        _ => Ok(()),
+                    }
                 });
             if v.is_err() {
                 warn!("Could not remove files after preprocessing failed!");
@@ -503,6 +574,7 @@ where
 
                 if let Err(e) = storage
                     .put_preprocessor_cache_entry(&preprocessor_key, preprocessor_cache_entry)
+                    .await
                 {
                     debug!("Failed to update preprocessor cache: {}", e);
                 }
@@ -555,7 +627,7 @@ const INCBIN_DIRECTIVE: &[u8] = b".incbin";
 fn process_preprocessed_file(
     input_file: &Path,
     cwd: &Path,
-    bytes: &mut Vec<u8>,
+    bytes: &mut [u8],
     included_files: &mut HashMap<PathBuf, String>,
     config: PreprocessorCacheModeConfig,
     time_of_compilation: std::time::SystemTime,
@@ -1490,6 +1562,7 @@ mod test {
         t("mm", Language::ObjectiveCxx);
 
         t("cu", Language::Cuda);
+        t("hip", Language::Hip);
     }
 
     #[test]
