@@ -17,7 +17,6 @@ use directories::ProjectDirs;
 use fs::File;
 use fs_err as fs;
 use once_cell::sync::Lazy;
-use regex::Regex;
 #[cfg(any(feature = "dist-client", feature = "dist-server"))]
 use serde::ser::Serializer;
 use serde::{
@@ -79,20 +78,19 @@ fn default_toolchain_cache_size() -> u64 {
 }
 
 pub fn parse_size(val: &str) -> Option<u64> {
-    let re = Regex::new(r"^(\d+)([KMGT])$").expect("Fixed regex parse failure");
-    re.captures(val)
-        .and_then(|caps| {
-            caps.get(1)
-                .and_then(|size| u64::from_str(size.as_str()).ok())
-                .map(|size| (size, caps.get(2)))
-        })
-        .and_then(|(size, suffix)| match suffix.map(|s| s.as_str()) {
-            Some("K") => Some(1024 * size),
-            Some("M") => Some(1024 * 1024 * size),
-            Some("G") => Some(1024 * 1024 * 1024 * size),
-            Some("T") => Some(1024 * 1024 * 1024 * 1024 * size),
-            _ => None,
-        })
+    let multiplier = match val.chars().last() {
+        Some('K') => 1024,
+        Some('M') => 1024 * 1024,
+        Some('G') => 1024 * 1024 * 1024,
+        Some('T') => 1024 * 1024 * 1024 * 1024,
+        _ => 1,
+    };
+    let val = if multiplier > 1 && !val.is_empty() {
+        val.split_at(val.len() - 1).0
+    } else {
+        val
+    };
+    u64::from_str(val).ok().map(|size| size * multiplier)
 }
 
 #[cfg(any(feature = "dist-client", feature = "dist-server"))]
@@ -224,15 +222,31 @@ pub struct GHACacheConfig {
 /// Please change this value freely if we have a better choice.
 const DEFAULT_MEMCACHED_CACHE_EXPIRATION: u32 = 86400;
 
-#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+fn default_memcached_cache_expiration() -> u32 {
+    DEFAULT_MEMCACHED_CACHE_EXPIRATION
+}
+
+#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Default)]
 #[serde(deny_unknown_fields)]
 pub struct MemcachedCacheConfig {
+    #[serde(alias = "endpoint")]
     pub url: String,
+
+    /// Username to authenticate with.
+    pub username: Option<String>,
+
+    /// Password to authenticate with.
+    pub password: Option<String>,
+
     /// the expiration time in seconds.
     ///
     /// Default to 24 hours (86400)
     /// Up to 30 days (2592000)
+    #[serde(default = "default_memcached_cache_expiration")]
     pub expiration: u32,
+
+    #[serde(default)]
+    pub key_prefix: String,
 }
 
 /// redis has no default TTL - all caches live forever
@@ -241,21 +255,49 @@ pub struct MemcachedCacheConfig {
 ///
 /// Please change this value freely if we have a better choice.
 const DEFAULT_REDIS_CACHE_TTL: u64 = 0;
-#[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
+pub const DEFAULT_REDIS_DB: u32 = 0;
+#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Default)]
 #[serde(deny_unknown_fields)]
 pub struct RedisCacheConfig {
-    pub url: String,
-    /// the ttl time in seconds.
+    /// The single-node redis endpoint.
+    /// Mutually exclusive with `cluster_endpoints`.
+    pub endpoint: Option<String>,
+
+    /// The redis cluster endpoints.
+    /// Mutually exclusive with `endpoint`.
+    pub cluster_endpoints: Option<String>,
+
+    /// Username to authenticate with.
+    pub username: Option<String>,
+
+    /// Password to authenticate with.
+    pub password: Option<String>,
+
+    /// The redis URL.
+    /// Deprecated in favor of `endpoint`.
+    pub url: Option<String>,
+
+    /// the db number to use
+    ///
+    /// Default to 0
+    #[serde(default)]
+    pub db: u32,
+
+    /// the ttl (expiration) time in seconds.
     ///
     /// Default to infinity (0)
-    #[serde(default)]
+    #[serde(default, alias = "expiration")]
     pub ttl: u64,
+
+    #[serde(default)]
+    pub key_prefix: String,
 }
 
 #[derive(Debug, PartialEq, Eq, Serialize, Deserialize)]
 #[serde(deny_unknown_fields)]
 pub struct WebdavCacheConfig {
     pub endpoint: String,
+    #[serde(default)]
     pub key_prefix: String,
     pub username: Option<String>,
     pub password: Option<String>,
@@ -267,6 +309,7 @@ pub struct WebdavCacheConfig {
 pub struct S3CacheConfig {
     pub bucket: String,
     pub region: Option<String>,
+    #[serde(default)]
     pub key_prefix: String,
     pub no_credentials: bool,
     pub endpoint: Option<String>,
@@ -278,6 +321,7 @@ pub struct S3CacheConfig {
 #[serde(deny_unknown_fields)]
 pub struct OSSCacheConfig {
     pub bucket: String,
+    #[serde(default)]
     pub key_prefix: String,
     pub endpoint: Option<String>,
     pub no_credentials: bool,
@@ -551,37 +595,51 @@ pub struct EnvConfig {
     cache: CacheConfigs,
 }
 
+fn key_prefix_from_env_var(env_var_name: &str) -> String {
+    env::var(env_var_name)
+        .ok()
+        .as_ref()
+        .map(|s| s.trim_end_matches('/'))
+        .filter(|s| !s.is_empty())
+        .unwrap_or_default()
+        .to_owned()
+}
+
+fn number_from_env_var<A: std::str::FromStr>(env_var_name: &str) -> Option<Result<A>>
+where
+    <A as FromStr>::Err: std::fmt::Debug,
+{
+    let value = env::var(env_var_name).ok()?;
+
+    value
+        .parse::<A>()
+        .map_err(|err| anyhow!("{env_var_name} value is invalid: {err:?}"))
+        .into()
+}
+
+fn bool_from_env_var(env_var_name: &str) -> Result<Option<bool>> {
+    env::var(env_var_name)
+        .ok()
+        .map(|value| match value.to_lowercase().as_str() {
+            "true" | "on" | "1" => Ok(true),
+            "false" | "off" | "0" => Ok(false),
+            _ => bail!(
+                "{} must be 'true', 'on', '1', 'false', 'off' or '0'.",
+                env_var_name
+            ),
+        })
+        .transpose()
+}
+
 fn config_from_env() -> Result<EnvConfig> {
     // ======= AWS =======
     let s3 = if let Ok(bucket) = env::var("SCCACHE_BUCKET") {
         let region = env::var("SCCACHE_REGION").ok();
-        let no_credentials =
-            env::var("SCCACHE_S3_NO_CREDENTIALS").map_or(Ok(false), |val| match val.as_str() {
-                "true" | "1" => Ok(true),
-                "false" | "0" => Ok(false),
-                _ => bail!("SCCACHE_S3_NO_CREDENTIALS must be 'true', '1', 'false', or '0'."),
-            })?;
-        let use_ssl = env::var("SCCACHE_S3_USE_SSL")
-            .ok()
-            .map(|value| value != "off");
-        let server_side_encryption =
-            env::var("SCCACHE_S3_SERVER_SIDE_ENCRYPTION")
-                .ok()
-                .map_or(Ok(Some(false)), |val| match val.as_str() {
-                    "true" | "1" => Ok(Some(true)),
-                    "false" | "0" => Ok(Some(false)),
-                    _ => bail!(
-                        "SCCACHE_S3_SERVER_SIDE_ENCRYPTION must be 'true', '1', 'false', or '0'."
-                    ),
-                })?;
+        let no_credentials = bool_from_env_var("SCCACHE_S3_NO_CREDENTIALS")?.unwrap_or(false);
+        let use_ssl = bool_from_env_var("SCCACHE_S3_USE_SSL")?;
+        let server_side_encryption = bool_from_env_var("SCCACHE_S3_SERVER_SIDE_ENCRYPTION")?;
         let endpoint = env::var("SCCACHE_ENDPOINT").ok();
-        let key_prefix = env::var("SCCACHE_S3_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .map(|s| s.to_owned() + "/")
-            .unwrap_or_default();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_S3_KEY_PREFIX");
 
         Some(S3CacheConfig {
             bucket,
@@ -604,27 +662,75 @@ fn config_from_env() -> Result<EnvConfig> {
     }
 
     // ======= redis =======
-    let ttl = match env::var("SCCACHE_REDIS_TTL") {
-        Ok(v) => v
-            .parse()
-            .map_err(|err| anyhow!("SCCACHE_REDIS_TTL value is invalid: {err:?}"))?,
-        Err(_) => DEFAULT_REDIS_CACHE_TTL,
+    let redis = match (
+        env::var("SCCACHE_REDIS").ok(),
+        env::var("SCCACHE_REDIS_ENDPOINT").ok(),
+        env::var("SCCACHE_REDIS_CLUSTER_ENDPOINTS").ok(),
+    ) {
+        (None, None, None) => None,
+        (url, endpoint, cluster_endpoints) => {
+            let db = number_from_env_var("SCCACHE_REDIS_DB")
+                .transpose()?
+                .unwrap_or(DEFAULT_REDIS_DB);
+
+            let username = env::var("SCCACHE_REDIS_USERNAME").ok();
+            let password = env::var("SCCACHE_REDIS_PASSWORD").ok();
+
+            let ttl = number_from_env_var("SCCACHE_REDIS_EXPIRATION")
+                .or_else(|| number_from_env_var("SCCACHE_REDIS_TTL"))
+                .transpose()?
+                .unwrap_or(DEFAULT_REDIS_CACHE_TTL);
+
+            let key_prefix = key_prefix_from_env_var("SCCACHE_REDIS_KEY_PREFIX");
+
+            Some(RedisCacheConfig {
+                url,
+                endpoint,
+                cluster_endpoints,
+                username,
+                password,
+                db,
+                ttl,
+                key_prefix,
+            })
+        }
     };
-    let redis = env::var("SCCACHE_REDIS")
-        .ok()
-        .map(|url| RedisCacheConfig { url, ttl });
+
+    if env::var_os("SCCACHE_REDIS_EXPIRATION").is_some()
+        && env::var_os("SCCACHE_REDIS_TTL").is_some()
+    {
+        bail!("You mustn't set both SCCACHE_REDIS_EXPIRATION and SCCACHE_REDIS_TTL. Use only one.");
+    }
 
     // ======= memcached =======
-    let expiration = match env::var("SCCACHE_MEMCACHED_EXPIRATION").ok() {
-        None => DEFAULT_MEMCACHED_CACHE_EXPIRATION,
-        Some(v) => v
-            .parse()
-            .map_err(|err| anyhow!("SCCACHE_MEMCACHED_EXPIRATION value is invalid: {err:?}"))?,
+    let memcached = if let Ok(url) =
+        env::var("SCCACHE_MEMCACHED").or_else(|_| env::var("SCCACHE_MEMCACHED_ENDPOINT"))
+    {
+        let username = env::var("SCCACHE_MEMCACHED_USERNAME").ok();
+        let password = env::var("SCCACHE_MEMCACHED_PASSWORD").ok();
+
+        let expiration = number_from_env_var("SCCACHE_MEMCACHED_EXPIRATION")
+            .transpose()?
+            .unwrap_or(DEFAULT_MEMCACHED_CACHE_EXPIRATION);
+
+        let key_prefix = key_prefix_from_env_var("SCCACHE_MEMCACHED_KEY_PREFIX");
+
+        Some(MemcachedCacheConfig {
+            url,
+            username,
+            password,
+            expiration,
+            key_prefix,
+        })
+    } else {
+        None
     };
 
-    let memcached = env::var("SCCACHE_MEMCACHED")
-        .ok()
-        .map(|url| MemcachedCacheConfig { url, expiration });
+    if env::var_os("SCCACHE_MEMCACHED").is_some()
+        && env::var_os("SCCACHE_MEMCACHED_ENDPOINT").is_some()
+    {
+        bail!("You mustn't set both SCCACHE_MEMCACHED and SCCACHE_MEMCACHED_ENDPOINT. Please, use only SCCACHE_MEMCACHED_ENDPOINT.");
+    }
 
     // ======= GCP/GCS =======
     if (env::var("SCCACHE_GCS_CREDENTIALS_URL").is_ok()
@@ -638,15 +744,7 @@ fn config_from_env() -> Result<EnvConfig> {
     }
 
     let gcs = env::var("SCCACHE_GCS_BUCKET").ok().map(|bucket| {
-        let key_prefix = env::var("SCCACHE_GCS_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .unwrap_or_default()
-            .to_owned();
-
-
+        let key_prefix = key_prefix_from_env_var("SCCACHE_GCS_KEY_PREFIX");
 
         if env::var("SCCACHE_GCS_OAUTH_URL").is_ok() {
             eprintln!("SCCACHE_GCS_OAUTH_URL has been deprecated");
@@ -664,11 +762,11 @@ fn config_from_env() -> Result<EnvConfig> {
             // TODO: unsure if these should warn during the configuration loading
             // or at the time when they're actually used to connect to GCS
             Ok(_) => {
-                warn!("Invalid SCCACHE_GCS_RW_MODE-- defaulting to READ_ONLY.");
+                warn!("Invalid SCCACHE_GCS_RW_MODE -- defaulting to READ_ONLY.");
                 CacheModeConfig::ReadOnly
             }
             _ => {
-                warn!("No SCCACHE_GCS_RW_MODE specified-- defaulting to READ_ONLY.");
+                warn!("No SCCACHE_GCS_RW_MODE specified -- defaulting to READ_ONLY.");
                 CacheModeConfig::ReadOnly
             }
         };
@@ -691,17 +789,13 @@ fn config_from_env() -> Result<EnvConfig> {
             enabled: true,
             version,
         })
-    } else if let Ok(enabled) = env::var("SCCACHE_GHA_ENABLED") {
-        // If only SCCACHE_GHA_ENABLED has been set, enable with
+    } else if bool_from_env_var("SCCACHE_GHA_ENABLED")?.unwrap_or(false) {
+        // If only SCCACHE_GHA_ENABLED has been set to the true value, enable with
         // default version.
-        if enabled == "on" || enabled == "true" {
-            Some(GHACacheConfig {
-                enabled: true,
-                version: "".to_string(),
-            })
-        } else {
-            None
-        }
+        Some(GHACacheConfig {
+            enabled: true,
+            version: "".to_string(),
+        })
     } else {
         None
     };
@@ -711,13 +805,7 @@ fn config_from_env() -> Result<EnvConfig> {
         env::var("SCCACHE_AZURE_CONNECTION_STRING"),
         env::var("SCCACHE_AZURE_BLOB_CONTAINER"),
     ) {
-        let key_prefix = env::var("SCCACHE_AZURE_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .unwrap_or_default()
-            .to_owned();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_AZURE_KEY_PREFIX");
         Some(AzureCacheConfig {
             connection_string,
             container,
@@ -729,13 +817,7 @@ fn config_from_env() -> Result<EnvConfig> {
 
     // ======= WebDAV =======
     let webdav = if let Ok(endpoint) = env::var("SCCACHE_WEBDAV_ENDPOINT") {
-        let key_prefix = env::var("SCCACHE_WEBDAV_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .unwrap_or_default()
-            .to_owned();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_WEBDAV_KEY_PREFIX");
         let username = env::var("SCCACHE_WEBDAV_USERNAME").ok();
         let password = env::var("SCCACHE_WEBDAV_PASSWORD").ok();
         let token = env::var("SCCACHE_WEBDAV_TOKEN").ok();
@@ -754,20 +836,9 @@ fn config_from_env() -> Result<EnvConfig> {
     // ======= OSS =======
     let oss = if let Ok(bucket) = env::var("SCCACHE_OSS_BUCKET") {
         let endpoint = env::var("SCCACHE_OSS_ENDPOINT").ok();
-        let key_prefix = env::var("SCCACHE_OSS_KEY_PREFIX")
-            .ok()
-            .as_ref()
-            .map(|s| s.trim_end_matches('/'))
-            .filter(|s| !s.is_empty())
-            .unwrap_or_default()
-            .to_owned();
+        let key_prefix = key_prefix_from_env_var("SCCACHE_OSS_KEY_PREFIX");
 
-        let no_credentials =
-            env::var("SCCACHE_OSS_NO_CREDENTIALS").map_or(Ok(false), |val| match val.as_str() {
-                "true" | "1" => Ok(true),
-                "false" | "0" => Ok(false),
-                _ => bail!("SCCACHE_OSS_NO_CREDENTIALS must be 'true', '1', 'false', or '0'."),
-            })?;
+        let no_credentials = bool_from_env_var("SCCACHE_OSS_NO_CREDENTIALS")?.unwrap_or(false);
 
         Some(OSSCacheConfig {
             bucket,
@@ -796,16 +867,11 @@ fn config_from_env() -> Result<EnvConfig> {
         .and_then(|v| parse_size(&v));
 
     let mut preprocessor_mode_config = PreprocessorCacheModeConfig::activated();
-    let preprocessor_mode_overridden = match env::var("SCCACHE_DIRECT").as_deref() {
-        Ok("on") | Ok("true") => {
-            preprocessor_mode_config.use_preprocessor_cache_mode = true;
-            true
-        }
-        Ok("off") | Ok("false") => {
-            preprocessor_mode_config.use_preprocessor_cache_mode = false;
-            true
-        }
-        _ => false,
+    let preprocessor_mode_overridden = if let Some(value) = bool_from_env_var("SCCACHE_DIRECT")? {
+        preprocessor_mode_config.use_preprocessor_cache_mode = value;
+        true
+    } else {
+        false
     };
 
     let (disk_rw_mode, disk_rw_mode_overridden) = match env::var("SCCACHE_LOCAL_RW_MODE")
@@ -815,7 +881,7 @@ fn config_from_env() -> Result<EnvConfig> {
         Ok("READ_ONLY") => (CacheModeConfig::ReadOnly, true),
         Ok("READ_WRITE") => (CacheModeConfig::ReadWrite, true),
         Ok(_) => {
-            warn!("Invalid SCCACHE_LOCAL_RW_MODE-- defaulting to READ_WRITE.");
+            warn!("Invalid SCCACHE_LOCAL_RW_MODE -- defaulting to READ_WRITE.");
             (CacheModeConfig::ReadWrite, false)
         }
         _ => (CacheModeConfig::ReadWrite, false),
@@ -1155,7 +1221,8 @@ pub mod server {
 #[test]
 fn test_parse_size() {
     assert_eq!(None, parse_size(""));
-    assert_eq!(None, parse_size("100"));
+    assert_eq!(None, parse_size("bogus value"));
+    assert_eq!(Some(100), parse_size("100"));
     assert_eq!(Some(2048), parse_size("2K"));
     assert_eq!(Some(10 * 1024 * 1024), parse_size("10M"));
     assert_eq!(Some(TEN_GIGS), parse_size("10G"));
@@ -1178,8 +1245,13 @@ fn config_overrides() {
                 rw_mode: CacheModeConfig::ReadWrite,
             }),
             redis: Some(RedisCacheConfig {
-                url: "myotherredisurl".to_owned(),
+                endpoint: Some("myotherredisurl".to_owned()),
                 ttl: 24 * 3600,
+                key_prefix: "/redis/prefix".into(),
+                db: 10,
+                username: Some("user".to_owned()),
+                password: Some("secret".to_owned()),
+                ..Default::default()
             }),
             ..Default::default()
         },
@@ -1196,10 +1268,14 @@ fn config_overrides() {
             memcached: Some(MemcachedCacheConfig {
                 url: "memurl".to_owned(),
                 expiration: 24 * 3600,
+                key_prefix: String::new(),
+                ..Default::default()
             }),
             redis: Some(RedisCacheConfig {
-                url: "myredisurl".to_owned(),
-                ttl: 24 * 3600,
+                url: Some("myredisurl".to_owned()),
+                ttl: 25 * 3600,
+                key_prefix: String::new(),
+                ..Default::default()
             }),
             ..Default::default()
         },
@@ -1211,8 +1287,13 @@ fn config_overrides() {
         Config::from_env_and_file_configs(env_conf, file_conf),
         Config {
             cache: Some(CacheType::Redis(RedisCacheConfig {
-                url: "myotherredisurl".to_owned(),
+                endpoint: Some("myotherredisurl".to_owned()),
                 ttl: 24 * 3600,
+                key_prefix: "/redis/prefix".into(),
+                db: 10,
+                username: Some("user".to_owned()),
+                password: Some("secret".to_owned()),
+                ..Default::default()
             }),),
             fallback_cache: DiskCacheConfig {
                 dir: "/env-cache".into(),
@@ -1254,7 +1335,7 @@ fn test_s3_no_credentials_invalid() {
 
     let error = config_from_env().unwrap_err();
     assert_eq!(
-        "SCCACHE_S3_NO_CREDENTIALS must be 'true', '1', 'false', or '0'.",
+        "SCCACHE_S3_NO_CREDENTIALS must be 'true', 'on', '1', 'false', 'off' or '0'.",
         error.to_string()
     );
 
@@ -1373,12 +1454,24 @@ enabled = true
 version = "sccache"
 
 [cache.memcached]
-url = "..."
-expiration = 86400
+# Deprecated alias for `endpoint`
+# url = "127.0.0.1:11211"
+endpoint = "tcp://127.0.0.1:11211"
+# Username and password for authentication
+username = "user"
+password = "passwd"
+expiration = 90000
+key_prefix = "/custom/prefix/if/need"
 
 [cache.redis]
-url = "redis://user:passwd@1.2.3.4:6379/1"
-ttl = 86400
+url = "redis://user:passwd@1.2.3.4:6379/?db=1"
+endpoint = "redis://127.0.0.1:6379"
+cluster_endpoints = "tcp://10.0.0.1:6379,redis://10.0.0.2:6379"
+username = "another_user"
+password = "new_passwd"
+db = 12
+expiration = 86400
+key_prefix = "/my/redis/cache"
 
 [cache.s3]
 bucket = "name"
@@ -1428,12 +1521,21 @@ no_credentials = true
                     version: "sccache".to_string()
                 }),
                 redis: Some(RedisCacheConfig {
-                    url: "redis://user:passwd@1.2.3.4:6379/1".to_owned(),
+                    url: Some("redis://user:passwd@1.2.3.4:6379/?db=1".to_owned()),
+                    endpoint: Some("redis://127.0.0.1:6379".to_owned()),
+                    cluster_endpoints: Some("tcp://10.0.0.1:6379,redis://10.0.0.2:6379".to_owned()),
+                    username: Some("another_user".to_owned()),
+                    password: Some("new_passwd".to_owned()),
+                    db: 12,
                     ttl: 24 * 3600,
+                    key_prefix: "/my/redis/cache".into(),
                 }),
                 memcached: Some(MemcachedCacheConfig {
-                    url: "...".to_owned(),
-                    expiration: 24 * 3600
+                    url: "tcp://127.0.0.1:11211".to_owned(),
+                    username: Some("user".to_owned()),
+                    password: Some("passwd".to_owned()),
+                    expiration: 25 * 3600,
+                    key_prefix: "/custom/prefix/if/need".into(),
                 }),
                 s3: Some(S3CacheConfig {
                     bucket: "name".to_owned(),
