@@ -1,14 +1,14 @@
+use fs_err as fs;
 #[cfg(any(feature = "dist-client", feature = "dist-server"))]
 use sccache::config::HTTPUrl;
 use sccache::dist::{self, SchedulerStatusResult, ServerId};
 use sccache::server::ServerInfo;
 use std::env;
-use std::fs;
 use std::io::Write;
 use std::net::{self, IpAddr, SocketAddr};
 use std::path::{Path, PathBuf};
 use std::process::{Command, Output, Stdio};
-use std::str;
+use std::str::{self, FromStr};
 use std::thread;
 use std::time::{Duration, Instant};
 
@@ -25,20 +25,10 @@ use predicates::prelude::*;
 use serde::Serialize;
 use uuid::Uuid;
 
-#[cfg(feature = "dist-server")]
-macro_rules! matches {
-    ($expression:expr, $($pattern:tt)+) => {
-        match $expression {
-            $($pattern)+ => true,
-            _ => false
-        }
-    }
-}
-
 const CONTAINER_NAME_PREFIX: &str = "sccache_dist_test";
 const DIST_IMAGE: &str = "sccache_dist_test_image";
 const DIST_DOCKERFILE: &str = include_str!("Dockerfile.sccache-dist");
-const DIST_IMAGE_BWRAP_PATH: &str = "/bwrap";
+const DIST_IMAGE_BWRAP_PATH: &str = "/usr/bin/bwrap";
 const MAX_STARTUP_WAIT: Duration = Duration::from_secs(5);
 
 const DIST_SERVER_TOKEN: &str = "THIS IS THE TEST TOKEN";
@@ -53,33 +43,44 @@ const TC_CACHE_SIZE: u64 = 1024 * 1024 * 1024; // 1 gig
 pub fn start_local_daemon(cfg_path: &Path, cached_cfg_path: &Path) {
     // Don't run this with run() because on Windows `wait_with_output`
     // will hang because the internal server process is not detached.
-    let _ = sccache_command()
+    if !sccache_command()
         .arg("--start-server")
+        // Uncomment following lines to debug locally.
+        // .env("SCCACHE_LOG", "debug")
+        // .env(
+        //     "SCCACHE_ERROR_LOG",
+        //     env::temp_dir().join("sccache_local_daemon.txt"),
+        // )
         .env("SCCACHE_CONF", cfg_path)
         .env("SCCACHE_CACHED_CONF", cached_cfg_path)
         .status()
         .unwrap()
-        .success();
+        .success()
+    {
+        panic!("Failed to start local daemon");
+    }
 }
-pub fn stop_local_daemon() {
+
+pub fn stop_local_daemon() -> bool {
     trace!("sccache --stop-server");
-    drop(
-        sccache_command()
-            .arg("--stop-server")
-            .stdout(Stdio::null())
-            .stderr(Stdio::null())
-            .status(),
-    );
+    sccache_command()
+        .arg("--stop-server")
+        .stdout(Stdio::null())
+        .stderr(Stdio::null())
+        .status()
+        .map_or(false, |status| status.success())
 }
 
 pub fn get_stats<F: 'static + Fn(ServerInfo)>(f: F) {
     sccache_command()
-        .args(&["--show-stats", "--stats-format=json"])
+        .args(["--show-stats", "--stats-format=json"])
         .assert()
         .success()
         .stdout(predicate::function(move |output: &[u8]| {
             let s = str::from_utf8(output).expect("Output not UTF-8");
-            f(serde_json::from_str(s).expect("Failed to parse JSON stats"));
+            let stats = serde_json::from_str(s).expect("Failed to parse JSON stats");
+            eprintln!("get server stats: {stats:?}");
+            f(stats);
             true
         }));
 }
@@ -98,13 +99,13 @@ pub fn zero_stats() {
 
 pub fn write_json_cfg<T: Serialize>(path: &Path, filename: &str, contents: &T) {
     let p = path.join(filename);
-    let mut f = fs::File::create(&p).unwrap();
+    let mut f = fs::File::create(p).unwrap();
     f.write_all(&serde_json::to_vec(contents).unwrap()).unwrap();
 }
 
 pub fn write_source(path: &Path, filename: &str, contents: &str) {
     let p = path.join(filename);
-    let mut f = fs::File::create(&p).unwrap();
+    let mut f = fs::File::create(p).unwrap();
     f.write_all(contents.as_bytes()).unwrap();
 }
 
@@ -126,7 +127,10 @@ pub fn sccache_dist_path() -> PathBuf {
     assert_cmd::cargo::cargo_bin("sccache-dist")
 }
 
-pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
+pub fn sccache_client_cfg(
+    tmpdir: &Path,
+    preprocessor_cache_mode: bool,
+) -> sccache::config::FileConfig {
     let cache_relpath = "client-cache";
     let dist_cache_relpath = "client-dist-cache";
     fs::create_dir(tmpdir.join(cache_relpath)).unwrap();
@@ -134,6 +138,10 @@ pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
 
     let disk_cache = sccache::config::DiskCacheConfig {
         dir: tmpdir.join(cache_relpath),
+        preprocessor_cache_mode: sccache::config::PreprocessorCacheModeConfig {
+            use_preprocessor_cache_mode: preprocessor_cache_mode,
+            ..Default::default()
+        },
         ..Default::default()
     };
     sccache::config::FileConfig {
@@ -145,6 +153,8 @@ pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
             memcached: None,
             redis: None,
             s3: None,
+            webdav: None,
+            oss: None,
         },
         dist: sccache::config::DistConfig {
             auth: Default::default(), // dangerously_insecure
@@ -157,6 +167,7 @@ pub fn sccache_client_cfg(tmpdir: &Path) -> sccache::config::FileConfig {
         server_startup_timeout_ms: None,
     }
 }
+
 #[cfg(feature = "dist-server")]
 fn sccache_scheduler_cfg() -> sccache::config::scheduler::Config {
     sccache::config::scheduler::Config {
@@ -167,6 +178,7 @@ fn sccache_scheduler_cfg() -> sccache::config::scheduler::Config {
         },
     }
 }
+
 #[cfg(feature = "dist-server")]
 fn sccache_server_cfg(
     tmpdir: &Path,
@@ -268,9 +280,11 @@ impl DistSystem {
                 "-e",
                 "SCCACHE_NO_DAEMON=1",
                 "-e",
-                "SCCACHE_LOG=sccache=trace",
+                "SCCACHE_LOG=debug",
                 "-e",
                 "RUST_BACKTRACE=1",
+                "--network",
+                "host",
                 "-v",
                 &format!("{}:/sccache-dist", self.sccache_dist.to_str().unwrap()),
                 "-v",
@@ -334,9 +348,11 @@ impl DistSystem {
                 "--name",
                 &server_name,
                 "-e",
-                "SCCACHE_LOG=sccache=trace",
+                "SCCACHE_LOG=debug",
                 "-e",
                 "RUST_BACKTRACE=1",
+                "--network",
+                "host",
                 "-v",
                 &format!("{}:/sccache-dist", self.sccache_dist.to_str().unwrap()),
                 "-v",
@@ -364,7 +380,7 @@ impl DistSystem {
 
         check_output(&output);
 
-        let server_ip = self.container_ip(&server_name);
+        let server_ip = IpAddr::from_str("127.0.0.1").unwrap();
         let server_cfg = sccache_server_cfg(&self.tmpdir, self.scheduler_url(), server_ip);
         fs::File::create(&server_cfg_path)
             .unwrap()
@@ -388,7 +404,7 @@ impl DistSystem {
         handler: S,
     ) -> ServerHandle {
         let server_addr = {
-            let ip = self.host_interface_ip();
+            let ip = IpAddr::from_str("127.0.0.1").unwrap();
             let listener = net::TcpListener::bind(SocketAddr::from((ip, 0))).unwrap();
             listener.local_addr().unwrap()
         };
@@ -404,7 +420,8 @@ impl DistSystem {
             ForkResult::Child => {
                 env::set_var("SCCACHE_LOG", "sccache=trace");
                 env_logger::try_init().unwrap();
-                void::unreachable(server.start().unwrap())
+                server.start().unwrap();
+                unreachable!();
             }
         };
 
@@ -461,8 +478,7 @@ impl DistSystem {
     }
 
     pub fn scheduler_url(&self) -> HTTPUrl {
-        let ip = self.container_ip(self.scheduler_name.as_ref().unwrap());
-        let url = format!("http://{}:{}", ip, SCHEDULER_PORT);
+        let url = format!("http://127.0.0.1:{}", SCHEDULER_PORT);
         HTTPUrl::from_url(reqwest::Url::parse(&url).unwrap())
     }
 
@@ -474,37 +490,6 @@ impl DistSystem {
         assert!(res.status().is_success());
         bincode::deserialize_from(res).unwrap()
     }
-
-    fn container_ip(&self, name: &str) -> IpAddr {
-        let output = Command::new("docker")
-            .args(&[
-                "inspect",
-                "--format",
-                "{{ .NetworkSettings.IPAddress }}",
-                name,
-            ])
-            .output()
-            .unwrap();
-        check_output(&output);
-        let stdout = String::from_utf8(output.stdout).unwrap();
-        stdout.trim().to_owned().parse().unwrap()
-    }
-
-    // The interface that the host sees on the docker network (typically 'docker0')
-    fn host_interface_ip(&self) -> IpAddr {
-        let output = Command::new("docker")
-            .args(&[
-                "inspect",
-                "--format",
-                "{{ .NetworkSettings.Gateway }}",
-                self.scheduler_name.as_ref().unwrap(),
-            ])
-            .output()
-            .unwrap();
-        check_output(&output);
-        let stdout = String::from_utf8(output.stdout).unwrap();
-        stdout.trim().to_owned().parse().unwrap()
-    }
 }
 
 // If you want containers to hang around (e.g. for debugging), comment out the "rm -f" lines
